@@ -41,6 +41,7 @@ except ImportError:
     from ansible.template import Templar
     from ansible.parsing.mod_args import ModuleArgsParser
     from ansible.plugins import module_loader
+    from ansible.errors import AnsibleParserError
     ANSIBLE_VERSION = 2
 
     def parse_yaml_from_file(filepath):
@@ -59,6 +60,7 @@ except ImportError:
         return templar.template(varname, **kwargs)
 
 LINE_NUMBER_KEY = '__line__'
+FILENAME_KEY = '__file__'
 
 VALID_KEYS = [
     'name', 'action', 'when', 'async', 'poll', 'notify',
@@ -99,11 +101,13 @@ def tokenize(line):
 
     args = list()
     kwargs = dict()
+    nonkvfound = False
     for arg in tokens[1:]:
-        if "=" in arg:
+        if "=" in arg and not nonkvfound:
             kv = arg.split("=", 1)
             kwargs[kv[0]] = kv[1]
         else:
+            nonkvfound = True
             args.append(arg)
     return (command, args, kwargs)
 
@@ -117,7 +121,7 @@ def _playbook_items(pb_data):
         return [item for play in pb_data for item in play.items()]
 
 
-def find_children(playbook):
+def find_children(playbook, playbook_dir):
     if not os.path.exists(playbook[0]):
         return []
     if playbook[1] == 'role':
@@ -131,7 +135,7 @@ def find_children(playbook):
     basedir = os.path.dirname(playbook[0])
     items = _playbook_items(playbook_ds)
     for item in items:
-        for child in play_children(basedir, item, playbook[1]):
+        for child in play_children(basedir, item, playbook[1], playbook_dir):
             if "$" in child['path'] or "{{" in child['path']:
                 continue
             valid_tokens = list()
@@ -159,7 +163,7 @@ def template(basedir, value, vars, fail_on_undefined=False, **kwargs):
     return value
 
 
-def play_children(basedir, item, parent_type):
+def play_children(basedir, item, parent_type, playbook_dir):
     delegate_map = {
         'tasks': _taskshandlers_children,
         'pre_tasks': _taskshandlers_children,
@@ -273,7 +277,7 @@ def rolename(filepath):
 
 def _kv_to_dict(v):
     (command, args, kwargs) = tokenize(v)
-    return (dict(module=command, module_arguments=args, **kwargs))
+    return (dict(__ansible_module__=command, __ansible_arguments__=args, **kwargs))
 
 
 def normalize_task_v2(task):
@@ -281,12 +285,24 @@ def normalize_task_v2(task):
 
     result = dict()
     mod_arg_parser = ModuleArgsParser(task)
-    action, arguments, result['delegate_to'] = mod_arg_parser.parse()
+    try:
+        action, arguments, result['delegate_to'] = mod_arg_parser.parse()
+    except AnsibleParserError as e:
+        task_info = "%s:%s" % (task[FILENAME_KEY], task[LINE_NUMBER_KEY])
+        del task[FILENAME_KEY]
+        del task[LINE_NUMBER_KEY]
+        try:
+            import pprint
+            pp = pprint.PrettyPrinter(indent=2)
+            task_pprint = pp.pformat(task)
+        except ImportError:
+            task_pprint = task
+        raise SystemExit("Couldn't parse task at %s (%s)\n%s" % (task_info, e.message, task_pprint))
 
     # denormalize shell -> command conversion
-    if '_use_shell' in arguments:
+    if '_uses_shell' in arguments:
         action = 'shell'
-        del(arguments['_use_shell'])
+        del(arguments['_uses_shell'])
 
     for (k, v) in list(task.items()):
         if k in ('action', 'local_action', 'args', 'delegate_to') or k == action:
@@ -296,13 +312,13 @@ def normalize_task_v2(task):
         else:
             result[k] = v
 
-    result['action'] = dict(module=action)
+    result['action'] = dict(__ansible_module__=action)
 
     if '_raw_params' in arguments:
-        result['action']['module_arguments'] = arguments['_raw_params'].split()
+        result['action']['__ansible_arguments__'] = arguments['_raw_params'].split()
         del(arguments['_raw_params'])
     else:
-        result['action']['module_arguments'] = list()
+        result['action']['__ansible_arguments__'] = list()
     result['action'].update(arguments)
     return result
 
@@ -314,7 +330,7 @@ def normalize_task_v1(task):
             if k == 'local_action' or k == 'action':
                 if not isinstance(v, dict):
                     v = _kv_to_dict(v)
-                v['module_arguments'] = v.get('module_arguments', list())
+                v['__ansible_arguments__'] = v.get('__ansible_arguments__', list())
                 result['action'] = v
             else:
                 result[k] = v
@@ -322,10 +338,10 @@ def normalize_task_v1(task):
             if isinstance(v, basestring):
                 v = _kv_to_dict(k + ' ' + v)
             elif not v:
-                v = dict(module=k)
+                v = dict(__ansible_module__=k)
             else:
                 if isinstance(v, dict):
-                    v.update(dict(module=k))
+                    v.update(dict(__ansible_module__=k))
                 else:
                     if k == '__line__':
                         # Keep the line number stored
@@ -340,19 +356,28 @@ def normalize_task_v1(task):
                                            "Task: %s. Check the syntax of your playbook using "
                                            "ansible-playbook --syntax-check" %
                                            (str(v), type(v), k, str(task)))
-            v['module_arguments'] = v.get('module_arguments', list())
+            v['__ansible_arguments__'] = v.get('__ansible_arguments__', list())
             result['action'] = v
+    if 'module' in result['action']:
+        # this happens when a task uses
+        # local_action:
+        #   module: ec2
+        #   etc...
+        result['action']['__ansible_module__'] = result['action']['module']
+        del(result['action']['module'])
     if 'args' in result:
         result['action'].update(result.get('args'))
         del(result['args'])
     return result
 
 
-def normalize_task(task):
+def normalize_task(task, filename):
     if ANSIBLE_VERSION < 2:
-        return normalize_task_v1(task)
+        task = normalize_task_v1(task)
     else:
-        return normalize_task_v2(task)
+        task = normalize_task_v2(task)
+    task[FILENAME_KEY] = filename
+    return task
 
 
 def task_to_str(task):
@@ -361,9 +386,9 @@ def task_to_str(task):
         return name
     action = task.get("action")
     args = " " .join(["{0}={1}".format(k, v) for (k, v) in action.items()
-                     if k not in ["module", "module_arguments"]] +
-                     action.get("module_arguments"))
-    return "{0} {1}".format(action["module"], args)
+                     if k not in ["__ansible_module__", "__ansible_arguments__"]] +
+                     action.get("__ansible_arguments__"))
+    return "{0} {1}".format(action["__ansible_module__"], args)
 
 
 def extract_from_list(blocks, candidates):
@@ -393,16 +418,19 @@ def get_action_tasks(yaml, file):
     block_rescue_always = ('block', 'rescue', 'always')
     tasks[:] = [task for task in tasks if all(k not in task for k in block_rescue_always)]
 
-    return [normalize_task(task) for task in tasks
-            if 'include' not in task.keys()]
+    return [task for task in tasks if 'include' not in task.keys()]
 
 
-def parse_yaml_linenumbers(data):
+def get_normalized_tasks(yaml, file):
+    tasks = get_action_tasks(yaml, file)
+    return [normalize_task(task, file['path']) for task in tasks]
+
+
+def parse_yaml_linenumbers(data, filename):
     """Parses yaml as ansible.utils.parse_yaml but with linenumbers.
 
     The line numbers are stored in each node's LINE_NUMBER_KEY key.
     """
-    loader = yaml.Loader(data)
 
     def compose_node(parent, index):
         # the line number where the previous token has ended (plus empty lines)
@@ -416,7 +444,11 @@ def parse_yaml_linenumbers(data):
         mapping[LINE_NUMBER_KEY] = node.__line__
         return mapping
 
-    loader.compose_node = compose_node
-    loader.construct_mapping = construct_mapping
-    data = loader.get_single_data()
+    try:
+        loader = yaml.Loader(data)
+        loader.compose_node = compose_node
+        loader.construct_mapping = construct_mapping
+        data = loader.get_single_data()
+    except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
+        raise SystemExit("Failed to parse YAML in %s: %s" % (filename, str(e)))
     return data
